#!/usr/bin/python3
# coding:utf8
import argparse
import asyncio
import logging.config
import platform
import threading

import biliup.common.reload
from biliup.config import config
from biliup.downloader import check_url
from . import __version__, LOG_CONF
from .common.Daemon import Daemon
from .common.reload import AutoReload


def arg_parser():
    daemon = Daemon('watch_process.pid', lambda: main(args))
    parser = argparse.ArgumentParser(description='Stream download and upload, not only for bilibili.')
    parser.add_argument('--version', action='version', version=f"v{__version__}")
    parser.add_argument('-H', help='web api host [default: 0.0.0.0]', dest='host')
    parser.add_argument('-P', help='web api port [default: 19159]', default=19159, dest='port')
    parser.add_argument('--http', action='store_true', help='enable web api')
    parser.add_argument('--static-dir', help='web static files directory for custom ui')
    parser.add_argument('--password', help='web ui password ,default username is biliup', dest='password')
    parser.add_argument('-v', '--verbose', action="store_const", const=logging.DEBUG, help="Increase output verbosity")
    parser.add_argument('--config', type=argparse.FileType(mode='rb'),
                        help='Location of the configuration file (default "./config.yaml")')
    subparsers = parser.add_subparsers(help='Windows does not support this sub-command.')
    # create the parser for the "start" command
    parser_start = subparsers.add_parser('start', help='Run as a daemon process.')
    parser_start.set_defaults(func=daemon.start)
    parser_stop = subparsers.add_parser('stop', help='Stop daemon according to "watch_process.pid".')
    parser_stop.set_defaults(func=daemon.stop)
    parser_restart = subparsers.add_parser('restart')
    parser_restart.set_defaults(func=daemon.restart)
    parser.set_defaults(func=lambda: asyncio.run(main(args)))
    args = parser.parse_args()
    biliup.common.reload.program_args = args.__dict__
    if args.http:
        config.create_without_config_input(args.config)
    else:
        config.load(args.config)
    LOG_CONF.update(config.get('LOGGING', {}))
    if args.verbose:
        LOG_CONF['loggers']['biliup']['level'] = args.verbose
        LOG_CONF['root']['level'] = args.verbose
    logging.config.dictConfig(LOG_CONF)
    if platform.system() == 'Windows':
        return asyncio.run(main(args))
    args.func()


async def main(args):
    from .handler import event_manager

    event_manager.start()

    for plugin in event_manager.context['checker']:
        # 线程数量是固定的在开始运行的时候创建不会产生变化也不会闲置或者复用线程 所以无需使用线程池
        # 这里也无需使用异步方法 一个线程一个检测 异步方法让渡控制权没用
        threading.Thread(target=check_url, args=(event_manager.context['checker'][plugin],)).start()

    interval = config.get('check_sourcecode', 15)
    if args.http:
        import biliup.web
        runner, site = await biliup.web.service(args, event_manager)
        detector = AutoReload(event_manager, runner.cleanup, interval=interval)
        biliup.common.reload.global_reloader = detector
        await asyncio.gather(detector.astart(), site.start(), return_exceptions=True)
    else:
        # 模块更新自动重启
        detector = AutoReload(event_manager, interval=interval)
        await asyncio.gather(detector.astart(), return_exceptions=True)


if __name__ == '__main__':
    arg_parser()



import base64
import hashlib
import html
import json
import random
import time
from urllib.parse import parse_qs, unquote

import requests

from biliup.config import config
from biliup.plugins.Danmaku import DanmakuClient
from ..engine.decorators import Plugin
from ..engine.download import DownloadBase
from ..plugins import match1, logger


@Plugin.download(regexp=r'(?:https?://)?(?:(?:www|m)\.)?huya\.com')
class Huya(DownloadBase):
    def __init__(self, fname, url, suffix='flv'):
        super().__init__(fname, url, suffix)
        self.huya_danmaku = config.get('huya_danmaku', False)
        self.fake_headers['User-Agent'] = 'Mozilla/5.0 (Linux; Android 10; SM-G981B) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/80.0.3987.162 Mobile Safari/537.36'

    def check_stream(self, is_check=False):
        try:
            room_id = self.url.split('huya.com/')[1].split('/')[0].split('?')[0]
        except:
            logger.warning(f"{Huya.__name__}: {self.url}: 直播间地址错误")
            return False
        try:
            res = requests.get(f'https://m.huya.com/{room_id}', timeout=5, headers=self.fake_headers)
            res.close()
        except:
            logger.warning(f"{Huya.__name__}: {self.url}: 获取错误，本次跳过")
            return False

        if '"exceptionType":0' in res.text:
            logger.warning(f"{Huya.__name__}: {self.url}: 直播间地址错误")
            return False

        if '"eLiveStatus":2' not in res.text:
            # 没开播
            return False

        live_info = json.loads(res.text.split('"tLiveInfo":')[1].split(',"_classname":"LiveRoom.LiveInfo"}')[0] + '}')
        if live_info:
            try:
                # 最大录制码率
                huya_max_ratio = config.get('huya_max_ratio', 0)
                # 码率信息
                ratio_items = live_info['tLiveStreamInfo']['vBitRateInfo']['value']
                # 最大码率(不含hdr)
                max_ratio = live_info['iBitRate']
                # 录制码率
                record_ratio = 0
                # 如果限制了最大码率
                if huya_max_ratio != 0:
                    # 挑选合适的码率
                    for ratio_item in ratio_items:
                        # iBitRate = 0的就是最大码率
                        if ratio_item['iBitRate'] == 0:
                            ratio_item['iBitRate'] = max_ratio
                        # 不录制大于最大码率的
                        if huya_max_ratio >= ratio_item['iBitRate'] > record_ratio:
                            record_ratio = ratio_item['iBitRate']

                    # 原画
                    if record_ratio == max_ratio:
                        record_ratio = 0

                # 自选cdn
                huya_cdn = config.get('huyacdn', 'AL')
                # 流信息
                stream_items = live_info['tLiveStreamInfo']['vStreamInfo']['value']
                # 自选的流
                stream_selected = stream_items[0]
                for stream_item in stream_items:
                    if stream_item['sCdnType'] == huya_cdn:
                        stream_selected = stream_item
                        break

                url_query = parse_qs(stream_selected["sFlvAntiCode"])
                uid = random.randint(1400000000000, 1499999999999)
                ws_time = hex(int(time.time() + 21600))[2:]
                seq_id = round(time.time() * 1000) + uid
                ws_secret_prefix = base64.b64decode(unquote(url_query['fm'][0]).encode()).decode().split("_")[0]
                ws_secret_hash = hashlib.md5(
                    f'{seq_id}|{url_query["ctype"][0]}|{url_query["t"][0]}'.encode()).hexdigest()
                ws_secret = hashlib.md5(
                    f'{ws_secret_prefix}_{uid}_{stream_selected["sStreamName"]}_{ws_secret_hash}_{ws_time}'.encode()).hexdigest()

                self.room_title = live_info['sIntroduction']
                self.raw_stream_url = f'{stream_selected["sFlvUrl"]}/{stream_selected["sStreamName"]}.{stream_selected["sFlvUrlSuffix"]}?wsSecret={ws_secret}&wsTime={ws_time}&seqid={seq_id}&ctype={url_query["ctype"][0]}&ver=1&fs={url_query["fs"][0]}&t={url_query["t"][0]}&uid={uid}&ratio={record_ratio}'
                return True
            except:
                logger.warning(f"{Huya.__name__}: {self.url}: 解析错误")

        return False

    async def danmaku_download_start(self, filename):
        if self.huya_danmaku:
            logger.info("开始弹幕录制")
            self.danmaku = DanmakuClient(self.url, filename + "." + self.suffix)
            await self.danmaku.start()

    def close(self):
        if self.huya_danmaku:
            self.danmaku.stop()
            logger.info("结束弹幕录制")



import json
import random
from struct import pack, unpack
import aiohttp
import zlib


class Bilibili:
    wss_url = 'wss://broadcastlv.chat.bilibili.com/sub'
    heartbeat = b'\x00\x00\x00\x1f\x00\x10\x00\x01\x00\x00\x00\x02\x00\x00\x00\x01\x5b\x6f\x62\x6a\x65\x63\x74\x20' \
                b'\x4f\x62\x6a\x65\x63\x74\x5d '
    heartbeatInterval = 60

    @staticmethod
    async def get_ws_info(url):
        url = 'https://api.live.bilibili.com/room/v1/Room/room_init?id=' + url.split('/')[-1]
        reg_datas = []
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as resp:
                room_json = json.loads(await resp.text())
                room_id = room_json['data']['room_id']
                data = json.dumps({
                    'roomid': room_id,
                    'uid': int(1e14 + 2e14 * random.random()),
                    'protover': 1
                }, separators=(',', ':')).encode('ascii')
                data = (pack('>i', len(data) + 16) + b'\x00\x10\x00\x01' +
                        pack('>i', 7) + pack('>i', 1) + data)
                reg_datas.append(data)

        return Bilibili.wss_url, reg_datas

    @staticmethod
    def decode_msg(data):
        dm_list_compressed = []
        dm_list = []
        ops = []
        msgs = []
        while True:
            try:
                packetLen, headerLen, ver, op, seq = unpack('!IHHII', data[0:16])
            except Exception as e:
                break
            if len(data) < packetLen:
                break
            if ver == 1 or ver == 0:
                ops.append(op)
                dm_list.append(data[16:packetLen])
            elif ver == 2:
                dm_list_compressed.append(data[16:packetLen])
            if len(data) == packetLen:
                data = b''
                break
            else:
                data = data[packetLen:]

        for dm in dm_list_compressed:
            d = zlib.decompress(dm)
            while True:
                try:
                    packetLen, headerLen, ver, op, seq = unpack('!IHHII', d[0:16])
                except Exception as e:
                    break
                if len(d) < packetLen:
                    break
                ops.append(op)
                dm_list.append(d[16:packetLen])
                if len(d) == packetLen:
                    d = b''
                    break
                else:
                    d = d[packetLen:]

        for i, d in enumerate(dm_list):
            try:
                msg = {}
                if ops[i] == 5:
                    j = json.loads(d)
                    msg['msg_type'] = {
                        'SEND_GIFT': 'gift',
                        'DANMU_MSG': 'danmaku',
                        'WELCOME': 'enter',
                        'NOTICE_MSG': 'broadcast',
                        'LIVE_INTERACTIVE_GAME': 'interactive_danmaku'  # 新增互动弹幕，经测试与弹幕内容一致
                    }.get(j.get('cmd'), 'other')

                    # 2021-06-03 bilibili 字段更新, 形如 DANMU_MSG:4:0:2:2:2:0
                    if      msg.get('msg_type', 'UNKNOWN').startswith('DANMU_MSG'):
                        msg['msg_type'] = 'danmaku'

                    if      msg['msg_type'] == 'danmaku':
                        msg['name'] = (j.get('info', ['', '', ['', '']])[2][1] or
                                       j.get('data', {}).get('uname', ''))
                        msg['content'] = j.get('info', ['', ''])[1]
                        msg["color"] = f"{j.get('info', '16777215')[0][3]}"


                    elif    msg['msg_type'] == 'interactive_danmaku':
                        msg['name'] = j.get('data', {}).get('uname', '')
                        msg['content'] = j.get('data', {}).get('msg', '')
                        msg["color"] = '16777215'

                    elif    msg['msg_type'] == 'broadcast':
                        msg['type'] = j.get('msg_type', 0)
                        msg['roomid'] = j.get('real_roomid', 0)
                        msg['content'] = j.get('msg_common', '')
                        msg['raw'] = j
                    else:
                        msg['content'] = j
                else:
                    msg = {'name': '', 'content': d, 'msg_type': 'other'}
                msgs.append(msg)
            except Exception as Error:
                print(Error)

        return msgs



# 抖音的弹幕录制参考了 https://github.com/LyzenX/DouyinLiveRecorder 和 https://github.com/YunzhiYike/live-tool
# 2023.07.14：KNaiFen：这部分代码参考了https://github.com/SmallPeaches/DanmakuRender

import gzip
import time
import re
from urllib.parse import unquote

import requests
import json

from google.protobuf import json_format

from .douyin_util.dy_pb2 import PushFrame, Response, ChatMessage  # 反序列化后导入


class Douyin:
    headers = {
        'authority': 'live.douyin.com',
        'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.5112.81 Safari/537.36 Edg/104.0.1293.54',
    }
    heartbeat = b':\x02hb'
    heartbeatInterval = 10

    @staticmethod
    async def get_ws_info(url):
        web_rid = url.split('/')[-1]
        if not Douyin.headers.get('cookie'):
            response = requests.get(f'https://live.douyin.com/{web_rid}', headers=Douyin.headers, timeout=5)
            Douyin.headers['cookie'] = '__ac_nonce=' + response.cookies.get('__ac_nonce')
            response = requests.get(f'https://live.douyin.com/{web_rid}', headers=Douyin.headers, timeout=5)
            Douyin.headers['cookie'] += '; ttwid=' + response.cookies.get('ttwid')
        else:
            response = requests.get(f'https://live.douyin.com/{web_rid}', headers=Douyin.headers, timeout=5)

        render_data = re.findall(r"<script id=\"RENDER_DATA\" type=\"application/json\">.*?</script>", response.text)[0]
        render_data = unquote(render_data)
        render_data = re.sub(r'(<script.*?>|</script>)', '', render_data)
        data = json.loads(render_data)
        real_rid = data['app']['initialState']['roomStore']['roomInfo']['roomId']
        user_unique_id = data['app']['odin']['user_unique_id']

        url = f"wss://webcast3-ws-web-lf.douyin.com/webcast/im/push/v2/?app_name=douyin_web&version_code=180800&webcast_sdk_version=1.3.0&update_version_code=1.3.0&compress=gzip&internal_ext=internal_src:dim|wss_push_room_id:{real_rid}|wss_push_did:{user_unique_id}|dim_log_id:2023011316221327ACACF0E44A2C0E8200|fetch_time:${int(time.time())}123|seq:1|wss_info:0-1673598133900-0-0|wrds_kvs:WebcastRoomRankMessage-1673597852921055645_WebcastRoomStatsMessage-1673598128993068211&cursor=u-1_h-1_t-1672732684536_r-1_d-1&host=https://live.douyin.com&aid=6383&live_id=1&did_rule=3&debug=false&endpoint=live_pc&support_wrds=1&im_path=/webcast/im/fetch/&device_platform=web&cookie_enabled=true&screen_width=1228&screen_height=691&browser_language=zh-CN&browser_platform=Mozilla&browser_name=Mozilla&browser_version=5.0%20(Windows%20NT%2010.0;%20Win64;%20x64)%20AppleWebKit/537.36%20(KHTML,%20like%20Gecko)%20Chrome/100.0.4896.75%20Safari/537.36&browser_online=true&tz_name=Asia/Shanghai&identity=audience&room_id={real_rid}&heartbeatDuration=0&signature=00000000"
        return url, []

    @staticmethod
    def decode_msg(data):
        wssPackage = PushFrame()
        wssPackage.ParseFromString(data)
        logid = wssPackage.logId
        decompressed = gzip.decompress(wssPackage.payload)
        payloadPackage = Response()
        payloadPackage.ParseFromString(decompressed)

        ack = None
        if payloadPackage.needAck:
            obj = PushFrame()
            obj.payloadType = 'ack'
            obj.logId = logid
            obj.payloadType = payloadPackage.internalExt
            ack = obj.SerializeToString()

        msgs = []
        for msg in payloadPackage.messagesList:
            if msg.method == 'WebcastChatMessage':
                chatMessage = ChatMessage()
                chatMessage.ParseFromString(msg.payload)
                data = json_format.MessageToDict(chatMessage, preserving_proto_field_name=True)
                # name = data['user']['nickName']
                content = data['content']
                # msg_dict = {"time": now, "name": name, "content": content, "msg_type": "danmaku", "color": "ffffff"}
                msg_dict = {"content": content, "msg_type": "danmaku"}
                msgs.append(msg_dict)

        return msgs, ack



import copy
import logging
import subprocess
import time

from . import plugins
from .downloader import download
from .engine import invert_dict, Plugin
from biliup.config import config
from .engine.event import Event, EventManager
from .uploader import upload
from .engine.upload import UploadBase

TO_MODIFY = 'to_modify'
DOWNLOAD = 'download'
UPLOAD = 'upload'
logger = logging.getLogger('biliup')


def create_event_manager():
    streamer_url = {k: v['url'] for k, v in config['streamers'].items()}
    inverted_index = invert_dict(streamer_url)
    urls = list(inverted_index.keys())
    pool1_size = config.get('pool1_size', 3)
    pool2_size = config.get('pool2_size', 3)
    # 初始化事件管理器
    app = EventManager(config, pool1_size=pool1_size, pool2_size=pool2_size)
    app.context['urls'] = urls
    app.context['url_status'] = dict.fromkeys(inverted_index, 0)
    app.context['url_upload_count'] = dict.fromkeys(inverted_index, 0)
    # 正在上传的文件 用于同时上传一个url的时候过滤掉正在上传的
    app.context['upload_filename'] = []
    app.context['checker'] = Plugin(plugins).sorted_checker(urls)
    app.context['inverted_index'] = inverted_index
    app.context['streamer_url'] = streamer_url
    return app


event_manager = create_event_manager()


@event_manager.register(DOWNLOAD, block='Asynchronous1')
def process(name, url):
    stream_info = {
        'name': name,
        'url': url,
    }

    start_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
    if config['streamers'].get(name, {}).get('preprocessor'):
        processor(config['streamers'].get(name, {}).get('preprocessor'), f'{{"name": "{name}", "url": "{url}", "start_time": "{start_time}"}}')

    url_status = event_manager.context['url_status']
    # 下载开始
    url_status[url] = 1
    try:
        kwargs: dict = config['streamers'][name].copy()
        kwargs.pop('url')
        suffix = kwargs.get('format')
        if suffix:
            kwargs['suffix'] = suffix
        stream_info = download(name, url, **kwargs)


        video_list = [file.video for file in UploadBase.file_list(stream_info['name'])]

        if config['streamers'].get(name, {}).get('downloaded_processor'):
            processor(config['streamers'].get(name, {}).get('downloaded_processor'),
                f'{{"name": "{name}", "url": "{url}", "room_title": "{stream_info.get("title", "")}", "start_time": "{start_time}", "end_time": "{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}", "file_list": "{video_list}"}}')
    finally:
        # 下载结束
        url_status[url] = 0
        yield Event(UPLOAD, (stream_info,))


@event_manager.register(UPLOAD, block='Asynchronous2')
def process_upload(stream_info):
    url = stream_info['url']
    url_upload_count = event_manager.context['url_upload_count']
    # 上传开始
    url_upload_count[url] += 1
    try:
        upload(stream_info)
    finally:
        # 上传结束
        url_upload_count[url] -= 1


@event_manager.server()
class KernelFunc:
    def __init__(self, urls, url_status: dict, url_upload_count: dict, checker, inverted_index, streamer_url):
        self.urls = urls
        # 录制状态 0等待录制 1正在录制 2正在上传(废弃)
        self.url_status = url_status
        # 上传状态 0未上传 >=1正在上传
        self.url_upload_count = url_upload_count
        self.checker = checker
        self.inverted_index = inverted_index
        self.streamer_url = streamer_url

    @event_manager.register(TO_MODIFY)
    def modify(self, url):
        if not url:
            # ?????
            logger.debug('无人直播')
            return
        name = self.inverted_index[url]
        logger.debug(f'{name} 刚刚开播，去下载')
        return Event(DOWNLOAD, args=(name, url))

    def get_url_status(self):
        # 这里是为webui准备的
        # webui fix
        url_status = copy.deepcopy(self.url_status)

        # 上传的情况下修改status 2
        for url in self.url_upload_count:
            if self.url_upload_count[url] > 0:
                url_status[url] = 2

        return url_status


def processor(processors, data):
    for processor in processors:
        if processor.get('run'):
            try:
                process_output = subprocess.check_output(
                    processor['run'], shell=True,
                    input=data,
                    stderr=subprocess.STDOUT, text=True)
                logger.info(process_output.rstrip())
            except subprocess.CalledProcessError as e:
                logger.exception(e.output)
                continue



import json
import urllib.request

import requests
import re
from . import logger
from biliup.config import config
from ..engine.decorators import Plugin
from ..engine.download import DownloadBase
from biliup.plugins.Danmaku import DanmakuClient


@Plugin.download(regexp=r'(?:https?://)?(?:(?:www|m|live)\.)?douyin\.com')
class Douyin(DownloadBase):
    def __init__(self, fname, url, suffix='flv'):
        super().__init__(fname, url, suffix)
        self.douyin_danmaku = config.get('douyin_danmaku', False)

    def check_stream(self, is_check=False):
        douyin_url = "https://live.douyin.com/"
        headers = {
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) "
                          "Chrome/94.0.4606.71 Safari/537.36 Edg/94.0.992.38",
            "referer": douyin_url,
            "cookie": config.get('user', {}).get('douyin_cookie')
        }
        if len(self.url.split(douyin_url)) < 2:
            if len(self.url.split("douyin.com/user/")) < 2:
                logger.debug("直播间地址错误")
                return False
            else:
                mainPage = requests.get(self.url, headers=headers).text \
                    .split('<script id="RENDER_DATA" type="application/json">')[1].split('</script>')[0]
                txt = urllib.request.unquote(mainPage)
                rex = re.compile(r'(?<=\"web_rid\":\")[0-9]*(?=\")')
                rid = rex.findall(txt)[0]
        else:
            # 判断是否为纯数字房间号
            rid = self.url.split(douyin_url)[1]
            rid = '+{}'.format(rid) if rid.isdigit() else rid
        try:
            r1 = requests.get(douyin_url + rid, headers=headers).text \
                .split('<script id="RENDER_DATA" type="application/json">')[1].split('</script>')[0]
            r2 = urllib.request.unquote(r1)
            room_info = json.loads(r2)['app']['initialState']['roomStore']['roomInfo']['room']
        except:
            logger.warning("抖音 " + rid + "：获取错误，本次跳过")
            return False
        if room_info.get('status') != 2:
            logger.debug("主播未开播")
            return False
        if room_info.get('stream_url'):
            r5 = room_info['stream_url']['live_core_sdk_data']['pull_data']['stream_data']
            stream_data = json.loads(r5)['data']

            # 原画origin 蓝光uhd 超清hd 高清sd 标清ld 流畅md 仅音频ao
            quality_items = ['origin', 'uhd', 'hd', 'sd', 'ld', 'md']
            quality = config.get('douyin_quality', 'origin')
            if quality not in quality_items:
                quality = quality_items[0]

            # 如果没有这个画质则取相近的 优先低清晰度
            if quality not in stream_data:
                # 可选的清晰度 含自身
                optional_quality_items = [x for x in quality_items if x in stream_data.keys() or x == quality]
                # 自身在可选清晰度的位置
                optional_quality_index = optional_quality_items.index(quality)
                # 自身在所有清晰度的位置
                quality_index = quality_items.index(quality)
                # 高清晰度偏移
                quality_left_offset = None
                # 低清晰度偏移
                quality_right_offset = None

                if optional_quality_index + 1 < len(optional_quality_items):
                    quality_right_offset = quality_items.index(optional_quality_items[optional_quality_index + 1]) - quality_index

                if optional_quality_index - 1 >= 0:
                    quality_left_offset = quality_index - quality_items.index(optional_quality_items[optional_quality_index - 1])

                # 取相邻的清晰度
                if quality_right_offset <= quality_left_offset:
                    quality = optional_quality_items[optional_quality_index + 1]
                else:
                    quality = optional_quality_items[optional_quality_index - 1]



            self.raw_stream_url = json.loads(r5)['data'][quality]['main']['flv']
            self.room_title = room_info['title']
            return True

    async def danmaku_download_start(self, filename):
        if self.douyin_danmaku:
            logger.info("开始弹幕录制")
            self.danmaku = DanmakuClient(self.url, filename + "." + self.suffix)
            await self.danmaku.start()

    def close(self):
        if self.douyin_danmaku:
            self.danmaku.stop()
            logger.info("结束弹幕录制")



import random
import re
import subprocess
import time
from typing import Generator, List
from urllib.parse import urlencode

import requests
import yt_dlp

from . import logger
from ..engine.decorators import Plugin
from ..engine.download import DownloadBase
from biliup.config import config
from biliup.plugins.Danmaku import DanmakuClient

VALID_URL_BASE = r'(?:https?://)?(?:(?:www|go|m)\.)?twitch\.tv/(?P<id>[0-9_a-zA-Z]+)'
VALID_URL_VIDEOS = r'https?://(?:(?:www|go|m)\.)?twitch\.tv/(?P<id>[^/]+)/(?:videos|profile|clips)'
_CLIENT_ID = 'kimne78kx3ncx6brgo4mv6wki5h1ko'

# Twitch 授权信息是否到期
AUTH_EXPIRE_STATUS = False


@Plugin.download(regexp=VALID_URL_VIDEOS)
class TwitchVideos(DownloadBase):
    def __init__(self, fname, url, suffix='mp4'):
        DownloadBase.__init__(self, fname, url, suffix=suffix)
        self.is_download = True

    def check_stream(self, is_check=False):
        # TODO 这里原本的批量检测是有问题的 先用yt_dlp实现 等待后续新增新的批量检测方式 后续这里的auth信息和直播一样采用twitch_cookie
        with yt_dlp.YoutubeDL({'download_archive': 'archive.txt'}) as ydl:
            try:
                info = ydl.extract_info(self.url, download=False, process=False)
                for entry in info['entries']:
                    if ydl.in_download_archive(entry):
                        continue
                    if not is_check:
                        download_info = ydl.extract_info(entry['url'], download=False)
                        self.room_title = download_info['title']
                        self.raw_stream_url = download_info['url']
                        thumbnails = download_info.get('thumbnails')
                        if type(thumbnails) is list and len(thumbnails) > 0:
                            self.live_cover_url = thumbnails[len(thumbnails) - 1].get('url')
                        ydl.record_download_archive(entry)
                    return True
            except:
                logger.warning(f"{self.url}：获取错误")
                return False
        return False



@Plugin.download(regexp=VALID_URL_BASE)
class Twitch(DownloadBase):
    def __init__(self, fname, url, suffix='flv'):
        DownloadBase.__init__(self, fname, url, suffix=suffix)
        self.twitch_danmaku = config.get('twitch_danmaku', False)
        self.twitch_disable_ads = config.get('twitch_disable_ads', True)
        self.proc = None

    def check_stream(self, is_check=False):
        channel_name = re.match(VALID_URL_BASE, self.url).group('id').lower()
        user = post_gql({
            "query": '''
                query query($channel_name:String!) {
                    user(login: $channel_name){
                        stream {
                            id
                            title
                            type
                            previewImageURL(width: 0,height: 0)
                            playbackAccessToken(
                                params: {
                                    platform: "web",
                                    playerBackend: "mediaplayer",
                                    playerType: "site"
                                }
                            ) {
                                signature
                                value             
                            }
                        }
                    }
                }
            ''',
            'variables': {'channel_name': channel_name}
        }).get('data', {}).get('user')
        if not user:
            logger.warning(f"{Twitch.__name__}: {self.url}: 获取错误")
            return False
        elif not user['stream'] or user['stream']['type'] != 'live':
            return False

        self.room_title = user['stream']['title']
        self.live_cover_url = user['stream']['previewImageURL']
        if is_check:
            return True

        if self.downloader == 'ffmpeg':
            port = random.randint(1025, 65535)
            stream_shell = [
                "streamlink",
                "--player-external-http",  # 为外部程序提供流媒体数据
                # "--twitch-disable-ads",                     # 去广告，去掉、跳过嵌入的广告流
                # "--twitch-disable-hosting",               # 该参数从5.0起已被禁用
                "--twitch-disable-reruns",  # 如果该频道正在重放回放，不打开流
                "--player-external-http-port", str(port),  # 对外部输出流的端口
                self.url, "best"  # 流链接
            ]
            if self.twitch_disable_ads:  # 去广告，去掉、跳过嵌入的广告流
                stream_shell.insert(1, "--twitch-disable-ads")

            twitch_cookie = config.get('user', {}).get('twitch_cookie')
            # 在设置且有效的情况下使用
            if twitch_cookie and not AUTH_EXPIRE_STATUS:
                stream_shell.insert(1, "--twitch-api-header=Authorization=OAuth " + twitch_cookie)

            self.proc = subprocess.Popen(stream_shell)
            self.raw_stream_url = f"http://localhost:{port}"
            i = 0
            while i < 5:
                if not (self.proc.poll() is None):
                    return False
                time.sleep(1)
                i += 1
            return True
        else:
            query = {
                "player": "twitchweb",
                "p": random.randint(1000000, 10000000),
                "allow_source": "true",
                "allow_audio_only": "true",
                "allow_spectre": "false",
                'fast_bread': "true",
                'sig': user.get('stream').get('playbackAccessToken').get('signature'),
                'token': user.get('stream').get('playbackAccessToken').get('value'),
            }
            self.raw_stream_url = f'https://usher.ttvnw.net/api/channel/hls/{channel_name}.m3u8?{urlencode(query)}'
            return True

    @staticmethod
    def batch_check(check_urls: List[str]) -> Generator[str, None, None]:
        ops = []
        for url in check_urls:
            channel_name = re.match(VALID_URL_BASE, url).group('id')
            op = {
                "query": '''
                    query query($login:String!) {
                        user(login: $login){
                            stream {
                              type
                            }
                        }
                    }
                ''',
                'variables': {'login': channel_name.lower()}
            }
            ops.append(op)
        gql = post_gql(ops)
        for index, data in enumerate(gql):
            user = data.get('data', {}).get('user')
            if not user:
                logger.warning(f"{Twitch.__name__}: {check_urls[index]}: 获取错误")
                continue
            elif not user['stream'] or user['stream']['type'] != 'live':
                continue
            yield check_urls[index]

    async def danmaku_download_start(self, filename):
        if self.twitch_danmaku:
            logger.info("开始弹幕录制")
            self.danmaku = DanmakuClient(self.url, filename + "." + self.suffix)
            await self.danmaku.start()

    def close(self):
        if self.twitch_danmaku:
            self.danmaku.stop()
            logger.info("结束弹幕录制")
        try:
            if self.proc is not None:
                self.proc.terminate()
        except:
            logger.exception(f'terminate {self.fname} failed')


def post_gql(ops):
    global AUTH_EXPIRE_STATUS
    headers = {
        'Content-Type': 'text/plain;charset=UTF-8',
        'Client-ID': _CLIENT_ID,
    }
    twitch_cookie = config.get('user', {}).get('twitch_cookie')
    if not AUTH_EXPIRE_STATUS and twitch_cookie:
        headers['Authorization'] = f'OAuth {twitch_cookie}'

    gql = requests.post(
        'https://gql.twitch.tv/gql',
        json=ops,
        headers=headers,
        timeout=15)
    gql.close()
    data = gql.json()

    if isinstance(data, dict) and data.get('error') == 'Unauthorized':
        AUTH_EXPIRE_STATUS = True
        logger.warning("Twitch Cookie已失效请及时更换，之后操作将忽略Twitch Cookie")
        return post_gql(ops)

    return data



import time
from urllib.parse import parse_qs

import jsengine
import requests
from ykdl.util.match import match1

from biliup.config import config
from biliup.plugins.Danmaku import DanmakuClient
from ..engine.decorators import Plugin
from ..engine.download import DownloadBase
from ..plugins import logger

MD5FUN = r'function md5(string){function RotateLeft(lValue,iShiftBits){return(lValue<<iShiftBits)|(lValue>>>(32-iShiftBits))}function AddUnsigned(lX,lY){var lX4,lY4,lX8,lY8,lResult;lX8=(lX&0x80000000);lY8=(lY&0x80000000);lX4=(lX&0x40000000);lY4=(lY&0x40000000);lResult=(lX&0x3FFFFFFF)+(lY&0x3FFFFFFF);if(lX4&lY4){return(lResult^0x80000000^lX8^lY8)}if(lX4|lY4){if(lResult&0x40000000){return(lResult^0xC0000000^lX8^lY8)}else{return(lResult^0x40000000^lX8^lY8)}}else{return(lResult^lX8^lY8)}}function F(x,y,z){return(x&y)|((~x)&z)}function G(x,y,z){return(x&z)|(y&(~z))}function H(x,y,z){return(x^y^z)}function I(x,y,z){return(y^(x|(~z)))}function FF(a,b,c,d,x,s,ac){a=AddUnsigned(a,AddUnsigned(AddUnsigned(F(b,c,d),x),ac));return AddUnsigned(RotateLeft(a,s),b)};function GG(a,b,c,d,x,s,ac){a=AddUnsigned(a,AddUnsigned(AddUnsigned(G(b,c,d),x),ac));return AddUnsigned(RotateLeft(a,s),b)};function HH(a,b,c,d,x,s,ac){a=AddUnsigned(a,AddUnsigned(AddUnsigned(H(b,c,d),x),ac));return AddUnsigned(RotateLeft(a,s),b)};function II(a,b,c,d,x,s,ac){a=AddUnsigned(a,AddUnsigned(AddUnsigned(I(b,c,d),x),ac));return AddUnsigned(RotateLeft(a,s),b)};function ConvertToWordArray(string){var lWordCount;var lMessageLength=string.length;var lNumberOfWords_temp1=lMessageLength+8;var lNumberOfWords_temp2=(lNumberOfWords_temp1-(lNumberOfWords_temp1%64))/64;var lNumberOfWords=(lNumberOfWords_temp2+1)*16;var lWordArray=Array(lNumberOfWords-1);var lBytePosition=0;var lByteCount=0;while(lByteCount<lMessageLength){lWordCount=(lByteCount-(lByteCount%4))/4;lBytePosition=(lByteCount%4)*8;lWordArray[lWordCount]=(lWordArray[lWordCount]|(string.charCodeAt(lByteCount)<<lBytePosition));lByteCount++}lWordCount=(lByteCount-(lByteCount%4))/4;lBytePosition=(lByteCount%4)*8;lWordArray[lWordCount]=lWordArray[lWordCount]|(0x80<<lBytePosition);lWordArray[lNumberOfWords-2]=lMessageLength<<3;lWordArray[lNumberOfWords-1]=lMessageLength>>>29;return lWordArray};function WordToHex(lValue){var WordToHexValue="",WordToHexValue_temp="",lByte,lCount;for(lCount=0;lCount<=3;lCount++){lByte=(lValue>>>(lCount*8))&255;WordToHexValue_temp="0"+lByte.toString(16);WordToHexValue=WordToHexValue+WordToHexValue_temp.substr(WordToHexValue_temp.length-2,2)}return WordToHexValue};function Utf8Encode(string){string=string.replace(/\r\n/g,"\n");var utftext="";for(var n=0;n<string.length;n++){var c=string.charCodeAt(n);if(c<128){utftext+=String.fromCharCode(c)}else if((c>127)&&(c<2048)){utftext+=String.fromCharCode((c>>6)|192);utftext+=String.fromCharCode((c&63)|128)}else{utftext+=String.fromCharCode((c>>12)|224);utftext+=String.fromCharCode(((c>>6)&63)|128);utftext+=String.fromCharCode((c&63)|128)}}return utftext};var x=Array();var k,AA,BB,CC,DD,a,b,c,d;var S11=7,S12=12,S13=17,S14=22;var S21=5,S22=9,S23=14,S24=20;var S31=4,S32=11,S33=16,S34=23;var S41=6,S42=10,S43=15,S44=21;string=Utf8Encode(string);x=ConvertToWordArray(string);a=0x67452301;b=0xEFCDAB89;c=0x98BADCFE;d=0x10325476;for(k=0;k<x.length;k+=16){AA=a;BB=b;CC=c;DD=d;a=FF(a,b,c,d,x[k+0],S11,0xD76AA478);d=FF(d,a,b,c,x[k+1],S12,0xE8C7B756);c=FF(c,d,a,b,x[k+2],S13,0x242070DB);b=FF(b,c,d,a,x[k+3],S14,0xC1BDCEEE);a=FF(a,b,c,d,x[k+4],S11,0xF57C0FAF);d=FF(d,a,b,c,x[k+5],S12,0x4787C62A);c=FF(c,d,a,b,x[k+6],S13,0xA8304613);b=FF(b,c,d,a,x[k+7],S14,0xFD469501);a=FF(a,b,c,d,x[k+8],S11,0x698098D8);d=FF(d,a,b,c,x[k+9],S12,0x8B44F7AF);c=FF(c,d,a,b,x[k+10],S13,0xFFFF5BB1);b=FF(b,c,d,a,x[k+11],S14,0x895CD7BE);a=FF(a,b,c,d,x[k+12],S11,0x6B901122);d=FF(d,a,b,c,x[k+13],S12,0xFD987193);c=FF(c,d,a,b,x[k+14],S13,0xA679438E);b=FF(b,c,d,a,x[k+15],S14,0x49B40821);a=GG(a,b,c,d,x[k+1],S21,0xF61E2562);d=GG(d,a,b,c,x[k+6],S22,0xC040B340);c=GG(c,d,a,b,x[k+11],S23,0x265E5A51);b=GG(b,c,d,a,x[k+0],S24,0xE9B6C7AA);a=GG(a,b,c,d,x[k+5],S21,0xD62F105D);d=GG(d,a,b,c,x[k+10],S22,0x2441453);c=GG(c,d,a,b,x[k+15],S23,0xD8A1E681);b=GG(b,c,d,a,x[k+4],S24,0xE7D3FBC8);a=GG(a,b,c,d,x[k+9],S21,0x21E1CDE6);d=GG(d,a,b,c,x[k+14],S22,0xC33707D6);c=GG(c,d,a,b,x[k+3],S23,0xF4D50D87);b=GG(b,c,d,a,x[k+8],S24,0x455A14ED);a=GG(a,b,c,d,x[k+13],S21,0xA9E3E905);d=GG(d,a,b,c,x[k+2],S22,0xFCEFA3F8);c=GG(c,d,a,b,x[k+7],S23,0x676F02D9);b=GG(b,c,d,a,x[k+12],S24,0x8D2A4C8A);a=HH(a,b,c,d,x[k+5],S31,0xFFFA3942);d=HH(d,a,b,c,x[k+8],S32,0x8771F681);c=HH(c,d,a,b,x[k+11],S33,0x6D9D6122);b=HH(b,c,d,a,x[k+14],S34,0xFDE5380C);a=HH(a,b,c,d,x[k+1],S31,0xA4BEEA44);d=HH(d,a,b,c,x[k+4],S32,0x4BDECFA9);c=HH(c,d,a,b,x[k+7],S33,0xF6BB4B60);b=HH(b,c,d,a,x[k+10],S34,0xBEBFBC70);a=HH(a,b,c,d,x[k+13],S31,0x289B7EC6);d=HH(d,a,b,c,x[k+0],S32,0xEAA127FA);c=HH(c,d,a,b,x[k+3],S33,0xD4EF3085);b=HH(b,c,d,a,x[k+6],S34,0x4881D05);a=HH(a,b,c,d,x[k+9],S31,0xD9D4D039);d=HH(d,a,b,c,x[k+12],S32,0xE6DB99E5);c=HH(c,d,a,b,x[k+15],S33,0x1FA27CF8);b=HH(b,c,d,a,x[k+2],S34,0xC4AC5665);a=II(a,b,c,d,x[k+0],S41,0xF4292244);d=II(d,a,b,c,x[k+7],S42,0x432AFF97);c=II(c,d,a,b,x[k+14],S43,0xAB9423A7);b=II(b,c,d,a,x[k+5],S44,0xFC93A039);a=II(a,b,c,d,x[k+12],S41,0x655B59C3);d=II(d,a,b,c,x[k+3],S42,0x8F0CCC92);c=II(c,d,a,b,x[k+10],S43,0xFFEFF47D);b=II(b,c,d,a,x[k+1],S44,0x85845DD1);a=II(a,b,c,d,x[k+8],S41,0x6FA87E4F);d=II(d,a,b,c,x[k+15],S42,0xFE2CE6E0);c=II(c,d,a,b,x[k+6],S43,0xA3014314);b=II(b,c,d,a,x[k+13],S44,0x4E0811A1);a=II(a,b,c,d,x[k+4],S41,0xF7537E82);d=II(d,a,b,c,x[k+11],S42,0xBD3AF235);c=II(c,d,a,b,x[k+2],S43,0x2AD7D2BB);b=II(b,c,d,a,x[k+9],S44,0xEB86D391);a=AddUnsigned(a,AA);b=AddUnsigned(b,BB);c=AddUnsigned(c,CC);d=AddUnsigned(d,DD)}var temp=WordToHex(a)+WordToHex(b)+WordToHex(c)+WordToHex(d);return temp.toLowerCase()}'


@Plugin.download(regexp=r'(?:https?://)?(?:(?:www|m)\.)?douyu\.com')
class Douyu(DownloadBase):
    def __init__(self, fname, url, suffix='flv'):
        super().__init__(fname, url, suffix)
        self.douyu_danmaku = config.get('douyu_danmaku', False)

    def check_stream(self, is_check=False):
        if len(self.url.split("douyu.com/")) < 2:
            logger.warning(f"{Douyu.__name__}: {self.url}: 直播间地址错误")
            return False
        try:
            if 'm.douyu.com' in self.url:
                # 傻瓜化适配 m
                room_no = self.url.split('m.douyu.com/')[1].split('/')[0].split('?')[0]
                html = requests.get(f'https://www.douyu.com/{room_no}', headers=self.fake_headers).text
            else:
                html = requests.get(self.url, headers=self.fake_headers).text

            room_id = match1(html, r'\$ROOM\.room_id\s*=\s*(\d+)')
            show_status = int(match1(html, r'\$ROOM\.show_status\s*=\s*(\d+)'))
            video_loop = int(match1(html, r'"videoLoop"\s*:\s*(\d+)'))
            self.room_title = match1(html, r'<h3 class="Title-header">(.*?)</h3>')
            if not room_id:
                logger.warning(f"{Douyu.__name__}: {self.url}: 直播间不存在或已关闭")
                return False
            if show_status != 1:
                logger.debug(f"{Douyu.__name__}: {self.url}: 未开播")
                return False
            if video_loop != 0:
                logger.debug(f"{Douyu.__name__}: {self.url}: 正在放录播")
                return False
        except:
            logger.warning(f"{Douyu.__name__}: {self.url}: 获取直播间信息错误")
            return False

        if is_check:
            # 检测模式不获取流
            return True

        try:
            ctx = jsengine.jsengine()
            ub98484234_fun = match1(html, r'(var vdwdae325w_64we.+?)function k927cea2d4369').replace('return eval','return strc;')
            sign_fun = ctx.eval(f'{ub98484234_fun};ub98484234();').rstrip(';').replace('CryptoJS.MD5(cb).toString()','md5(cb)')
            sign_fun += f'("{room_id}","10000000000000000000000000001501","{int(time.time())}");{MD5FUN}'
            params = parse_qs(ctx.eval(sign_fun))
        except TypeError:
            logger.error(f"{Douyu.__name__}: {self.url}: 请安装至少一个 Javascript 解释器")
            return False
        except:
            logger.warning(f"{Douyu.__name__}: {self.url}: 获取签名参数异常")
            return False
        params['cdn'] = config.get('douyucdn', 'tct-h5')
        params['rate'] = config.get('douyu_rate', 0)

        try:
            live_data = self.get_play_info(room_id, params)
            if type(live_data) is not dict:
                return False
        except:
            logger.warning(f"{Douyu.__name__}: {self.url}: 获取下载信息错误")
            return False

        self.raw_stream_url = f"{live_data.get('rtmp_url')}/{live_data.get('rtmp_live')}"
        return True

    async def danmaku_download_start(self, filename):
        if self.douyu_danmaku:
            logger.info("开始弹幕录制")
            self.danmaku = DanmakuClient(self.url, filename + "." + self.suffix)
            await self.danmaku.start()

    def close(self):
        if self.douyu_danmaku:
            self.danmaku.stop()
            logger.info("结束弹幕录制")

    def get_play_info(self, room_id, params):
        live_data = requests.post(f'https://www.douyu.com/lapi/live/getH5Play/{room_id}', headers=self.fake_headers,
                                  params=params).json().get('data')
        if type(live_data) is dict:
            # 禁用斗鱼主线路
            if not live_data.get('rtmp_cdn', '').endswith('h5') or 'akm' in live_data.get('rtmp_cdn', ''):
                params['cdn'] = 'tct-h5'
                return self.get_play_info(room_id, params)
            return live_data

        return None



import asyncio
import logging
import os
import re
import subprocess
import sys
import threading
import time
from typing import Generator, List
from urllib.parse import urlparse

import requests
import stream_gears

from biliup.config import config

logger = logging.getLogger('biliup')


class DownloadBase:
    def __init__(self, fname, url, suffix=None, opt_args=None):
        self.danmaku = None
        self.room_title = None
        if opt_args is None:
            opt_args = []
        # 主播单独传参会覆盖全局设置。例如新增了一个全局的filename_prefix参数，在下面添加self.filename_prefix = config.get('filename_prefix'),
        # 即可通过self.filename_prefix在下载或者上传时候传递主播单独的设置参数用于调用（如果该主播有设置单独参数，将会优先使用单独参数；如无，则会优先你用全局参数。）
        self.fname = fname
        self.url = url
        self.suffix = suffix
        self.title = None
        self.live_cover_path = None
        self.downloader = config.get('downloader', 'stream-gears')
        # ffmpeg.exe -i  http://vfile1.grtn.cn/2018/1542/0254/3368/154202543368.ssm/154202543368.m3u8
        # -c copy -bsf:a aac_adtstoasc -movflags +faststart output.mp4
        self.raw_stream_url = None
        self.filename_prefix = config.get('filename_prefix')
        self.use_live_cover = config.get('use_live_cover', False)
        self.opt_args = opt_args
        # 是否是下载模式 跳过下播检测
        self.is_download = False
        self.live_cover_url = None
        self.fake_headers = {
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            'Accept-Encoding': 'gzip, deflate',
            'Accept-Language': 'zh-CN,zh;q=0.8,en-US;q=0.5,en;q=0.3',
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/92.0.4515.159 Safari/537.36',
        }

        self.default_output_args = [
            '-bsf:a', 'aac_adtstoasc',
        ]
        if config.get('segment_time'):
            self.default_output_args += \
                ['-to', f"{config.get('segment_time', '00:50:00')}"]
        else:
            self.default_output_args += \
                ['-fs', f"{config.get('file_size', '2621440000')}"]

    def check_stream(self, is_check=False):
        # is_check 是否是检测可以避免在检测是否可以录制的时候忽略一些耗时的操作
        logger.debug(self.fname, is_check)
        raise NotImplementedError()

    @staticmethod
    def batch_check(check_urls: List[str]) -> Generator[str, None, None]:
        # 批量检测直播或下载状态
        # 返回的是url_list
        raise NotImplementedError()

    def get_filename(self, is_fmt=False):
        if self.filename_prefix:  # 判断是否存在自定义录播命名设置
            filename = (self.filename_prefix.format(streamer=self.fname, title=self.room_title).encode(
                'unicode-escape').decode()).encode().decode("unicode-escape")
        else:
            filename = f'{self.fname}%Y-%m-%dT%H_%M_%S'
        filename = get_valid_filename(filename)
        if is_fmt:
            return time.strftime(filename.encode("unicode-escape").decode()).encode().decode("unicode-escape")
        else:
            return filename

    def download(self, filename):
        filename = self.get_filename()
        fmtname = time.strftime(filename.encode("unicode-escape").decode()).encode().decode("unicode-escape")

        threading.Thread(target=asyncio.run, args=(self.danmaku_download_start(fmtname),)).start()

        if self.downloader == 'streamlink':
            parsed_url = urlparse(self.raw_stream_url)
            path = parsed_url.path
            if '.flv' in path:  # streamlink无法处理flv,所以回退到ffmpeg
                return self.ffmpeg_download(fmtname)
            else:
                return self.streamlink_download(fmtname)
        elif self.downloader == 'ffmpeg':
            return self.ffmpeg_download(fmtname)

        stream_gears_download(self.raw_stream_url, self.fake_headers, filename, config.get('segment_time'),
                              config.get('file_size'))
        return True

    def streamlink_download(self, filename):  # streamlink+ffmpeg混合下载模式，适用于下载hls流
        streamlink_input_args = ['--stream-segment-threads', '3', '--hls-playlist-reload-attempts', '1']
        streamlink_cmd = ['streamlink', *streamlink_input_args, self.raw_stream_url, 'best', '-O']
        ffmpeg_input_args = ['-rw_timeout', '20000000']
        ffmpeg_cmd = ['ffmpeg', '-re', '-i', 'pipe:0', '-y', *ffmpeg_input_args, *self.default_output_args,
                      *self.opt_args, '-c', 'copy', '-f', self.suffix]
        # if config.get('segment_time'):
        #     ffmpeg_cmd += ['-f', 'segment',
        #              f'{filename} part-%03d.{self.suffix}']
        # else:
        #     ffmpeg_cmd += [
        #         f'{filename}.{self.suffix}.part']
        ffmpeg_cmd += [f'{filename}.{self.suffix}.part']
        streamlink_proc = subprocess.Popen(streamlink_cmd, stdout=subprocess.PIPE)
        ffmpeg_proc = subprocess.Popen(ffmpeg_cmd, stdin=streamlink_proc.stdout, stdout=subprocess.PIPE,
                                       stderr=subprocess.STDOUT)
        try:
            with ffmpeg_proc.stdout as stdout:
                for line in iter(stdout.readline, b''):
                    decode_line = line.decode(errors='ignore')
                    print(decode_line, end='', file=sys.stderr)
                    logger.debug(decode_line.rstrip())
            retval = ffmpeg_proc.wait()
        except KeyboardInterrupt:
            if sys.platform != 'win32':
                ffmpeg_proc.communicate(b'q')
            raise
        if retval != 0:
            return False
        return True

    def ffmpeg_download(self, filename):
        default_input_args = ['-headers', ''.join('%s: %s\r\n' % x for x in self.fake_headers.items()), '-rw_timeout',
                              '20000000']
        parsed_url = urlparse(self.raw_stream_url)
        path = parsed_url.path
        if '.m3u8' in path:
            default_input_args += ['-max_reload', '1000']
        args = ['ffmpeg', '-y', *default_input_args,
                '-i', self.raw_stream_url, *self.default_output_args, *self.opt_args,
                '-c', 'copy', '-f', self.suffix]
        # if config.get('segment_time'):
        #     args += ['-f', 'segment',
        #              f'{filename} part-%03d.{self.suffix}']
        # else:
        #     args += [
        #         f'{filename}.{self.suffix}.part']
        args += [f'{filename}.{self.suffix}.part']

        proc = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        try:
            with proc.stdout as stdout:
                for line in iter(stdout.readline, b''):  # b'\n'-separated lines
                    decode_line = line.decode(errors='ignore')
                    print(decode_line, end='', file=sys.stderr)
                    logger.debug(decode_line.rstrip())
            retval = proc.wait()
        except KeyboardInterrupt:
            if sys.platform != 'win32':
                proc.communicate(b'q')
            raise
        if retval != 0:
            return False
        return True

    async def danmaku_download_start(self, filename):
        pass

    def run(self):
        if not self.check_stream():
            return False
        file_name = self.file_name
        retval = self.download(file_name)
        logger.info(f'part: {file_name}.{self.suffix}')
        self.rename(f'{file_name}.{self.suffix}')
        return retval

    def start(self):
        logger.info(f'开始下载：{self.__class__.__name__} - {self.fname}')
        date = time.localtime()
        delay = int(config.get('delay', 0))
        # 重试次数
        retry_count = 0
        # delay 重试次数
        retry_count_delay = 0
        # delay 总重试次数 向上取整
        delay_all_retry_count = -(-delay // 60)

        while True:
            ret = False
            try:
                ret = self.run()
            except:
                logger.exception('Uncaught exception:')
            finally:
                self.close()
            if ret:
                if self.is_download:
                    # 成功下载后也不检测下一个需要下载的视频而是先上传等待下次检测保证上传时使用下载视频的标题
                    # 开启边录边传会快些
                    break
                # 成功下载重置重试次数
                retry_count = 0
                retry_count_delay = 0
            else:
                if retry_count < 3:
                    retry_count += 1
                    logger.info(f'获取流失败：{self.__class__.__name__} - {self.fname}，重试次数 {retry_count} / 3，等待 10 秒')
                    time.sleep(10)
                    continue
                if self.is_download:
                    # 下载模式如果下载失败重试三次后直接跳出
                    break

                if delay:
                    retry_count_delay += 1
                    if retry_count_delay > delay_all_retry_count:
                        logger.info(f'下播延迟检测结束：{self.__class__.__name__}:{self.fname}')
                        break
                    else:
                        if delay < 60:
                            logger.info(
                                f'下播延迟检测：{self.__class__.__name__} - {self.fname}，将在 {delay} 秒后检测开播状态')
                            time.sleep(delay)
                        else:
                            if retry_count_delay == 1:
                                # 只有第一次显示
                                logger.info(
                                    f'下播延迟检测：{self.__class__.__name__} - {self.fname}，每隔 60 秒检测开播状态，共检测 {delay_all_retry_count} 次')
                            time.sleep(60)
                        continue
                else:
                    break

        # 获取封面
        if self.use_live_cover and self.live_cover_url is not None:
            try:
                save_dir = f'cover/{self.__class__.__name__}/{self.fname}/'
                if not os.path.exists(save_dir):
                    os.makedirs(save_dir)
                fmtname = time.strftime(self.get_filename().encode("unicode-escape").decode(), date).encode().decode(
                    "unicode-escape")

                url_path = urlparse(self.live_cover_url).path
                suffix = None
                if '.jpg' in url_path:
                    suffix = 'jpg'
                elif '.png' in url_path:
                    suffix = 'png'

                if suffix:
                    live_cover_path = f'{save_dir}{fmtname}.{suffix}'
                    if os.path.exists(live_cover_path):
                        self.live_cover_path = live_cover_path
                    else:
                        response = requests.get(self.live_cover_url, headers=self.fake_headers, timeout=30)
                        with open(live_cover_path, 'wb') as f:
                            f.write(response.content)
                            f.close()
                            self.live_cover_path = live_cover_path
                    logger.info(
                        f'封面下载成功：{self.__class__.__name__} - {self.fname}：{os.path.abspath(self.live_cover_path)}')
                else:
                    logger.warning(f'封面下载失败：{self.__class__.__name__} - {self.fname}：封面格式不支持')

            except:
                logger.exception(f'封面下载失败：{self.__class__.__name__} - {self.fname}')

        logger.info(f'退出下载：{self.__class__.__name__} - {self.fname}')
        return {
            'name': self.fname,
            'url': self.url,
            'title': self.room_title,
            'date': date,
            'live_cover_path': self.live_cover_path,
            'is_download': self.is_download,
        }

    @staticmethod
    def rename(file_name):
        try:
            os.rename(file_name + '.part', file_name)
            logger.debug(f'更名 {file_name + ".part"} 为 {file_name}')
        except FileNotFoundError:
            logger.debug(f'FileNotFoundError: {file_name}')
        except FileExistsError:
            os.rename(file_name + '.part', file_name)
            logger.info(f'FileExistsError: 更名 {file_name + ".part"} 为 {file_name}')

    @property
    def file_name(self):
        if self.filename_prefix:  # 判断是否存在自定义录播命名设置
            filename = (self.filename_prefix.format(streamer=self.fname, title=self.room_title).encode(
                'unicode-escape').decode()).encode().decode("unicode-escape")
        else:
            filename = f'{self.fname}%Y-%m-%dT%H_%M_%S'
        filename = get_valid_filename(filename)
        return time.strftime(filename.encode("unicode-escape").decode()).encode().decode("unicode-escape")

    def close(self):
        pass


def stream_gears_download(url, headers, file_name, segment_time=None, file_size=None):
    class Segment:
        pass

    segment = Segment()
    if segment_time:
        seg_time = segment_time.split(':')
        print(int(seg_time[0]) * 60 * 60 + int(seg_time[1]) * 60 + int(seg_time[2]))
        segment.time = int(seg_time[0]) * 60 * 60 + int(seg_time[1]) * 60 + int(seg_time[2])
    if file_size:
        segment.size = file_size
    if file_size is None and segment_time is None:
        segment.size = 8 * 1024 * 1024 * 1024
    stream_gears.download(
        url,
        headers,
        file_name,
        segment
    )


def get_valid_filename(name):
    """
    Return the given string converted to a string that can be used for a clean
    filename. Remove leading and trailing spaces; convert other spaces to
    underscores; and remove anything that is not an alphanumeric, dash,
    underscore, or dot.
    # >>> get_valid_filename("john's portrait in 2004.jpg")
    >>> get_valid_filename("{self.fname}%Y-%m-%dT%H_%M_%S")
    '{self.fname}%Y-%m-%dT%H_%M_%S'
    """
    # s = str(name).strip().replace(" ", "_") #因为有些人会在主播名中间加入空格，为了避免和录播完毕自动改名冲突，所以注释掉
    s = re.sub(r"(?u)[^-\w.%{}\[\]【】「」\s]", "", str(name))
    if s in {"", ".", ".."}:
        raise RuntimeError("Could not derive file name from '%s'" % name)
    return s


# 抖音的弹幕录制参考了 https://github.com/LyzenX/DouyinLiveRecorder 和 https://github.com/YunzhiYike/live-tool

from datetime import datetime
import threading
import asyncio
import gzip
import re
import time
import re
import requests
import urllib
import json

import websocket
from google.protobuf import json_format

from .dy_pb2 import PushFrame, Response, ChatMessage



import asyncio
import random
import shutil
import subprocess
import threading

import yaml
import yt_dlp
import streamlink
import os
import time

from yt_dlp.utils import DateRange, UserNotLive
from biliup.config import config
from ..engine.decorators import Plugin
from ..engine.download import DownloadBase, get_valid_filename
from . import logger

VALID_URL_BASE = r'https?://(?:(?:www|m)\.)?youtube\.com/(?P<id>.*?)\??(.*?)'


@Plugin.download(regexp=VALID_URL_BASE)
class Youtube(DownloadBase):
    def __init__(self, fname, url, suffix='webm'):
        DownloadBase.__init__(self, fname, url, suffix=suffix)
        self.use_new_ytb_downloader = config.get('use_new_ytb_downloader', False)
        self.ytb_danmaku = config.get('ytb_danmaku', False)
        self.cookiejarFile = config.get('user', {}).get('youtube_cookie')
        self.vcodec = config.get('youtube_prefer_vcodec', 'av01|vp9|avc')
        self.acodec = config.get('youtube_prefer_acodec', 'opus|mp4a')
        self.resolution = config.get('youtube_max_resolution', '4320')
        self.filesize = config.get('youtube_max_videosize', '100G')
        self.beforedate = config.get('youtube_before_date', '20770707')
        self.afterdate = config.get('youtube_after_date', '19700101')
        # 需要下载的url
        self.download_url = None

    def check_stream(self, is_check=False):
        self.download_url = None
        if self.use_new_ytb_downloader and self.downloader == 'ffmpeg':
            _fname, self.room_title = self.get_stream_info(self.url)
            port = random.randint(1025, 65535)
            stream_shell = [
                "streamlink",
                "--player-external-http",  # 为外部程序提供流媒体数据
                "--player-external-http-port", str(port),  # 对外部输出流的端口
                self.url, "best"  # 流链接
            ]
            self.proc = subprocess.Popen(stream_shell)
            self.raw_stream_url = f"http://localhost:{port}"
            i = 0
            while i < 5:
                if not (self.proc.poll() is None):
                    return False
                time.sleep(1)
                i += 1
            return True
        else:
            with yt_dlp.YoutubeDL({'download_archive': 'archive.txt', 'cookiefile': self.cookiejarFile,
                                   'format': f"bestvideo[vcodec~='^({self.vcodec})'][height<={self.resolution}][filesize<{self.filesize}]+bestaudio[acodec~='^({self.acodec})']/best[height<={self.resolution}]/best",
                                   }) as ydl:
                try:
                    info = ydl.extract_info(self.url, download=False, process=False)
                    if type(info) is not dict:
                        logger.warning(f"{self.url}：获取错误")
                        return False
                except UserNotLive:
                    logger.warning(f"{self.url}：地址填写错误")
                    return False
                except:
                    logger.warning(f"{self.url}：获取错误")
                    return False

                if info.get('entries') is None:
                    if ydl.in_download_archive(info):
                        return False
                    # 视频取标题
                    self.room_title = info.get('title')
                    self.live_cover_url = info.get('thumbnail')
                    self.download_url = self.url
                    return True

                # 时间范围缓存避免每次都要读取
                cache_save_dir = './cache'
                cache_filename = f'{cache_save_dir}/yt_dlp_cache.yaml'
                if not os.path.exists(cache_save_dir):
                    os.makedirs(cache_save_dir)
                if not os.path.exists(cache_filename):
                    with open(cache_filename, 'w') as f:
                        f.close()

                cache = {}
                try:
                    with open(cache_filename, 'r', encoding='utf-8') as file:
                        cache = yaml.load(file, Loader=yaml.FullLoader)
                        if type(cache) is not dict:
                            cache = {}
                except:
                    logger.exception(f'读取{Youtube.__class__}缓存失败，将重置缓存')

                self.is_download = True

                for entry in info['entries']:
                    # 检测是否已下载
                    if ydl.in_download_archive(entry):
                        continue

                    try:
                        # 检测时间范围
                        if entry['id'] not in cache:
                            cache[entry['id']] = {}
                        upload_date = cache.get(entry['id']).get('upload_date')
                        if upload_date is None:
                            upload_date = ydl.extract_info(entry['url'], download=False).get('upload_date')

                        cache[entry['id']]['upload_date'] = upload_date
                        if upload_date not in DateRange(self.afterdate, self.beforedate):
                            continue
                    except:
                        continue

                    # 取 Playlist 内视频标题
                    self.room_title = entry['title']
                    thumbnails = entry.get('thumbnails', [])
                    if type(thumbnails) is list and len(thumbnails) > 0:
                        self.live_cover_url = thumbnails[len(thumbnails) - 1].get('url')
                    self.download_url = entry['url']
                    break

                with open(cache_filename, "w", encoding='utf-8') as f:
                    yaml.dump(cache, f, encoding='utf-8', allow_unicode=True)

                return self.download_url is not None

    def download(self, filename):
        if self.use_new_ytb_downloader and self.downloader == 'ffmpeg':
            if self.filename_prefix:  # 判断是否存在自定义录播命名设置
                filename = (self.filename_prefix.format(streamer=self.fname, title=self.room_title).encode(
                    'unicode-escape').decode()).encode().decode("unicode-escape")
            else:
                filename = f'{self.fname}%Y-%m-%dT%H_%M_%S'
            filename = get_valid_filename(filename)
            fmtname = time.strftime(filename.encode("unicode-escape").decode()).encode().decode("unicode-escape")
            threading.Thread(target=asyncio.run, args=(self.danmaku_download_start(fmtname),)).start()
            self.ffmpeg_download(fmtname)
        else:
            # ydl下载的文件在下载失败时不可控
            # 临时存储在其他地方
            if self.is_download:
                download_dir = f'./cache/youtube/{filename}'
            else:
                download_dir = '.'
            try:
                ydl_opts = {
                    'outtmpl': f'{download_dir}/{filename}.%(ext)s',
                    'format': f"bestvideo[vcodec~='^({self.vcodec})'][height<={self.resolution}][filesize<{self.filesize}]+bestaudio[acodec~='^({self.acodec})']/best[height<={self.resolution}]/best",
                    'cookiefile': self.cookiejarFile,
                    # 'proxy': proxyUrl,
                    'break_on_reject': True,
                }
                if self.is_download:
                    # 只有下载时记录已下载
                    ydl_opts['download_archive'] = 'archive.txt'

                with yt_dlp.YoutubeDL(ydl_opts) as ydl:
                    ydl.download([self.download_url])

                if self.is_download:
                    for file in os.listdir(download_dir):
                        shutil.move(f'{download_dir}/{file}', './')
            except:
                logger.exception(self.fname)
                return False
            finally:
                # 删除引用 让gc回收ydl释放文件句柄
                del ydl
                if self.is_download:
                    # 清理意外退出可能产生的多余文件
                    shutil.rmtree(download_dir)
            return True

    def get_stream_info(self, url):
        session = streamlink.Streamlink()
        # Streamlink 插件系统的内部运行机制，通过 URL 匹配到 YouTube 插件
        plugin = session.resolve_url(url)[1]
        streams = plugin(session=session, url=url)
        streams._get_streams()
        author = streams.get_author()
        title = streams.get_title()
        return author, title



import logging
import re
import time
from urllib.error import HTTPError

from .engine.decorators import Plugin

from .engine.download import DownloadBase
from .engine.event import Event
from .engine.upload import UploadBase
from .plugins import general
from biliup.config import config

logger = logging.getLogger('biliup')


def download(fname, url, **kwargs):
    pg = general.__plugin__(fname, url)
    for plugin in Plugin.download_plugins:
        if re.match(plugin.VALID_URL_BASE, url):
            pg = plugin(fname, url)
            for k in pg.__dict__:
                if kwargs.get(k):
                    pg.__dict__[k] = kwargs.get(k)
            break
    return pg.start()


def check_url(checker):
    from .handler import event_manager, TO_MODIFY, UPLOAD
    # 单主播检测延迟
    checker_sleep = config.get('checker_sleep', 15)
    # 平台检测延迟
    event_loop_interval = config.get('event_loop_interval', 40)
    context = event_manager.context
    class_reference = type(checker('', ''))
    while True:
        try:
            # 待检测url
            check_urls = []
            # 过滤url
            for url in checker.url_list:
                if context['url_status'][url] == 1:
                    logger.debug(f'{url}正在下载中，跳过检测')
                    continue

                # 检测之前可能未上传的视频
                if len(UploadBase.file_list(context['inverted_index'][url])) > 0:
                    event_manager.send_event(
                        Event(UPLOAD, args=({'name': context['inverted_index'][url], 'url': url},)))
                    # 这里不能直接修改url_upload_count 因为UPLOAD Event中有url_upload_count的修改
                    # 所以单独判断
                    if config.get('uploading_record', False):
                        logger.debug(f'{url}正在上传中，跳过检测')
                        continue
                else:
                    # 没有等待上传的视频
                    if context['url_upload_count'][url] > 0 and not config.get('uploading_record', False):
                        logger.debug(f'{url}正在上传中，跳过检测')
                        continue
                check_urls.append(url)

            if DownloadBase.batch_check != getattr(class_reference, DownloadBase.batch_check.__name__):
                # 如果支持批量检测
                # 发送下载的事件
                for url in class_reference.batch_check(check_urls):
                    event_manager.send_event(Event(TO_MODIFY, args=(url,)))
            else:
                # 不支持批量检测
                for (index, url) in enumerate(check_urls):
                    # 某个检测异常略过不应影响其他检测
                    try:
                        if index > 0:
                            logger.debug('歇息会')
                            time.sleep(checker_sleep)

                        if checker(context['inverted_index'][url], url).check_stream(True):
                            event_manager.send_event(Event(TO_MODIFY, args=(url,)))
                    except HTTPError as e:
                        logger.error(f'{checker.__module__} {e.url} => {e}')
                    except IOError:
                        logger.exception("IOError")
                    except:
                        logger.exception("Uncaught exception:")



        except:
            # 除了单个检测异常 其他异常会影响整体 直接略过本次 等待下次整体检测
            logger.exception("Uncaught exception:")

        time.sleep(event_loop_interval)



import re
import time
import requests

from biliup.config import config
from . import match1, logger
from biliup.plugins.Danmaku import DanmakuClient
from ..engine.decorators import Plugin
from ..engine.download import DownloadBase


@Plugin.download(regexp=r'(?:https?://)?(?:(?:www|m|live)\.)?bilibili\.com')
class Bilibili(DownloadBase):
    def __init__(self, fname, url, suffix='flv'):
        super().__init__(fname, url, suffix)
        self.fake_headers['Referer'] = 'https://live.bilibili.com'
        self.bilibili_danmaku = config.get('bilibili_danmaku', False)
        if config.get('user', {}).get('bili_cookie') is not None:
            self.fake_headers['cookie'] = config.get('user', {}).get('bili_cookie')

    def check_stream(self, is_check=False):

        # 预读配置
        params = {
            'room_id': match1(self.url, r'/(\d+)'),
            'protocol': '0,1',# 0: http_stream, 1: http_hls
            'format': '0,1,2',# 0: flv, 1: ts, 2: fmp4
            'codec': '0', # 0: avc, 1: hevc
            'qn': config.get('bili_qn', '10000'),
            'platform': 'html5', # 使用 html5 时默认屏蔽 p2p
            'dolby': '5',
            'panorama': '1'
        }
        protocol = config.get('bili_protocol', 'stream')
        perf_cdn = config.get('bili_perfCDN')
        bili_cdn_fallback = config.get('bili_cdn_fallback', True)
        force_source = config.get('bili_force_source', False)
        ov05_ip = config.get('bili_force_ov05_ip')
        bili_fallback_api = config.get('bili_fallback_api')
        cn01_domains = config.get('bili_force_cn01_domains', '').split(",")
        official_api_host = "https://api.live.bilibili.com"

        with requests.Session() as s:
            s.headers = self.fake_headers.copy()
            # 获取直播状态与房间标题
            info_by_room_url = f"{official_api_host}/xlive/web-room/v1/index/getInfoByRoom?room_id={params['room_id']}"
            try:
                room_info = s.get(info_by_room_url, timeout=3).json()
            except requests.exceptions.ConnectionError:
                logger.error(f"在连接到 {info_by_room_url} 时出现错误")
                return False
            if room_info['code'] != 0 or room_info['data']['room_info']['live_status'] != 1:
                logger.debug(room_info['message'])
                return False
            self.live_cover_url = room_info['data']['room_info']['cover']
            live_start_time = room_info['data']['room_info']['live_start_time']
            uname = room_info['data']['anchor_info']['base_info']['uname']
            if self.room_title is None:
                self.room_title = room_info['data']['room_info']['title']

            # 当 Cookie 存在，并且自定义APi使用Cookie开关关闭时，仅使用官方 Api
            isallow = True if s.headers.get('cookie') is None else config.get('user', {}).get('customAPI_use_cookie', False)
            try:
                play_info = get_play_info(s, isallow, official_api_host, params)
            except:
                logger.error("使用官方 Api 失败")
                return False
        if play_info['code'] != 0:
            logger.debug(play_info['message'])
            return False
        playurl_info = play_info['data']['playurl_info']['playurl']
        if playurl_info is None:
            logger.debug(play_info)
            logger.error('可能遇到地区限制，无法获取到直播流')
            return False
        streams = playurl_info['stream']
        stream = streams[1] if protocol.startswith('hls') and len(streams) > 1 else streams[0]
        if protocol == "hls_fmp4":
            if len(stream['format']) > 1:
                stream_info = stream['format'][1]['codec'][0]
            elif int(time.time()) - live_start_time <= 60:  # 等待60s，如果还没有fmp4流就回退到flv流
                return False
            else:
                if bili_fallback_api: #找不到fmp4流就自动回退到指定API请求flv流，适用于海外机下载
                    play_info = s.get(bili_fallback_api + '/xlive/web-room/v2/index/getRoomPlayInfo', params=params, timeout=5).json()
                    streams = play_info['data']['playurl_info']['playurl']['stream']
                stream = streams[0]
                stream_info = stream['format'][0]['codec'][0]
                logger.debug(f"获取{uname}房间fmp4流失败，回退到flv流")
        else:
            stream_info = stream['format'][0]['codec'][0]
        stream_url = {'base_url': stream_info['base_url'],}
        if perf_cdn is not None:
            perf_cdn_list = perf_cdn.split(',')
            for url_info in stream_info['url_info']:
                if 'host' in stream_url:
                    break
                for cdn in perf_cdn_list:
                    if cdn in url_info['extra']:
                        stream_url['host'] = url_info['host']
                        stream_url['extra'] = url_info['extra']
                        logger.debug(f"找到了perfCDN{stream_url['host']}")
                        break
        if len(stream_url) < 3:
            stream_url['host'] = stream_info['url_info'][-1]['host']
            stream_url['extra'] = stream_info['url_info'][-1]['extra']

        # 低级设置
        if "cn-gotcha01" in stream_url['extra']:
            # 强制替换cn-gotcha01的节点为指定节点 注意：只有大陆ip才能获取到cn-gotcha01的节点。
            if cn01_domains[0] != '':
                import random
                i = len(cn01_domains)
                while i:  # 测试节点是否可用
                    host = cn01_domains.pop(random.choice(range(i)))
                    i-=1
                    try:
                        if s.get(f"https://{host}{stream_url['base_url']}{stream_url['extra']}",
                                            stream=True).status_code == 200:  # 如果响应状态码是 200，跳出循环
                            stream_url['host'] = "https://" + host
                            logger.debug(f"节点 {host} 可用，替换为该节点")
                            break
                    except requests.exceptions.ConnectionError:  # 如果发生连接异常，继续下一次循环
                        logger.debug(f"节点 {host} 无法访问，尝试下一个节点。")
                        continue
                else:
                    logger.error("配置文件中的cn-gotcha01节点均不可用")
            # 强制去除 cn01线路的hls_ts与hls_fmp4流（beta）的 _bluray 文件名，从而实现获取真实原画流的目的
            if force_source:
                stream_url['base_url'] = re.sub(r'_bluray(?=(/index)?\.m3u8)', "", stream_url['base_url'], 1)
        self.raw_stream_url = stream_url['host'] + stream_url['base_url'] + stream_url['extra']

        # 强制替换ov05 302redirect之后的真实地址为指定的域名或ip达到自选ov05节点的目的
        if ov05_ip and "ov-gotcha05" in stream_url['host']:
            r = s.get(self.raw_stream_url, stream=True)
            self.raw_stream_url = re.sub(r".*(?=/d1--ov-gotcha05)", f"http://{ov05_ip}", r.url, 1)
            logger.debug(f"将ov-gotcha05的节点ip替换为了{ov05_ip}")

        if bili_cdn_fallback:
            try:
                if s.get(self.raw_stream_url, stream=True).status_code == 404:
                    stream_info['url_info'].reverse()
                    for url_info in stream_info['url_info']:
                        self.raw_stream_url = url_info['host'] + stream_info['base_url'] + url_info['extra']
                        if s.get(self.raw_stream_url, stream=True).status_code == 200:
                            break
            except Exception:
                pass
        return True

    async def danmaku_download_start(self, filename):
        if self.bilibili_danmaku:
            logger.info("开始弹幕录制")
            self.danmaku = DanmakuClient(self.url, filename + "." + self.suffix)
            await self.danmaku.start()

    def close(self):
        if self.bilibili_danmaku:
            self.danmaku.stop()
            logger.info("结束弹幕录制")
            # asyncio.run(self.danmaku.stop())


def get_play_info(s, isallow, official_api_host, params):
    if isallow:
        custom_api_host = \
            (lambda a: a if a.startswith(('http://', 'https://')) else 'http://' + a) \
            (config.get('bili_liveapi', official_api_host).rstrip('/'))
        try:
            return s.get(custom_api_host + '/xlive/web-room/v2/index/getRoomPlayInfo', params=params,
                         timeout=5).json()
        except requests.exceptions.ConnectionError:
            logger.error(f"{custom_api_host}连接失败，尝试回退至官方Api")
    return s.get(official_api_host + '/xlive/web-room/v2/index/getRoomPlayInfo', params=params, timeout=5).json()



import logging
import os
import shutil
import subprocess
from functools import reduce
from pathlib import Path
from typing import NamedTuple, Optional, List

from biliup.config import config

logger = logging.getLogger('biliup')


class UploadBase:
    class FileInfo(NamedTuple):
        video: str
        danmaku: Optional[str]

    def __init__(self, principal, data, persistence_path=None, postprocessor=None):
        self.principal = principal
        self.persistence_path = persistence_path
        self.data: dict = data
        self.post_processor = postprocessor

    @staticmethod
    def file_list(index) -> List[FileInfo]:
        from biliup.handler import event_manager
        media_extensions = ['.mp4', '.flv', '.3gp', '.webm', '.mkv', '.ts']

        # 获取文件列表
        file_list = []
        for file_name in os.listdir('.'):
            if index in file_name and os.path.isfile(file_name):
                file_list.append(file_name)
        file_list = sorted(file_list, key=lambda x: os.path.getctime(x))

        if len(file_list) == 0:
            return []

        # 正在上传的文件列表
        upload_filename: list = event_manager.context['upload_filename']

        results = []
        for index, file in enumerate(file_list):
            if file.endswith('.part'):
                new_name = os.path.splitext(file)[0]
                shutil.move(file, new_name)
                logger.info(f'{file}已更名为{new_name}')
                file_list[index] = new_name
                file = new_name

            name, ext = os.path.splitext(file)

            # 过滤不是视频的 如果是弹幕检测下弹幕存不存在
            if ext not in media_extensions:
                if ext == '.xml':
                    have_video = False
                    for media_extension in media_extensions:
                        if f"{name}{media_extension}" in file_list:
                            have_video = True
                            break
                    if not have_video:
                        logger.info(f'无视频，过滤删除-{file}')
                        UploadBase.remove_file(file)
                continue
            # 过滤正在上传的
            if file in upload_filename:
                continue

            file_size = os.path.getsize(file) / 1024 / 1024
            threshold = config.get('filtering_threshold', 0)
            if file_size <= threshold:
                logger.info(f'过滤删除-{file}')
                UploadBase.remove_file(file)
                continue

            video = file
            danmaku = None
            if f'{name}.xml' in file_list:
                danmaku = f'{name}.xml'

            result = UploadBase.FileInfo(video=video, danmaku=danmaku)

            results.append(result)

        return results

    @staticmethod
    def remove_filelist(file_list: List[FileInfo]):
        for f in file_list:
            UploadBase.remove_file(f.video)
            if f.danmaku is not None:
                UploadBase.remove_file(f.video)

    @staticmethod
    def remove_file(file: str):
        try:
            os.remove(file)
            logger.info(f'删除-{file}')
        except:
            logger.warning(f'删除失败-{file}')

    def upload(self, file_list: List[FileInfo]) -> List[FileInfo]:
        raise NotImplementedError()

    def start(self):
        file_list = UploadBase.file_list(self.principal)
        if len(file_list) > 0:
            video_list = [file.video for file in file_list]
            logger.info('准备上传' + self.data["format_title"])
            from biliup.handler import event_manager
            upload_filename: list = event_manager.context['upload_filename']
            try:
                event_manager.context['upload_filename'].extend(video_list)
                needed2process = self.upload(file_list)
                if needed2process:
                    self.postprocessor(needed2process)
            finally:
                event_manager.context['upload_filename'] = list(set(upload_filename) - set(video_list))

    def postprocessor(self, data: List[FileInfo]):
        # data = file_list
        if self.post_processor is None:
            # 删除封面
            if self.data.get('live_cover_path') is not None:
                UploadBase.remove_file(self.data['live_cover_path'])
            return self.remove_filelist(data)

        file_list = []
        for i in data:
            file_list.append(i.video)
            if i.danmaku is not None:
                file_list.append(i.danmaku)

        for post_processor in self.post_processor:
            if post_processor == 'rm':
                # 删除封面
                if self.data.get('live_cover_path') is not None:
                    UploadBase.remove_file(self.data['live_cover_path'])
                self.remove_filelist(data)
                continue
            if post_processor.get('mv'):
                for file in file_list:
                    path = Path(file)
                    dest = Path(post_processor['mv'])
                    if not dest.is_dir():
                        dest.mkdir(parents=True, exist_ok=True)
                    try:
                        shutil.move(path, dest / path.name)
                    except Exception as e:
                        logger.exception(e)
                        continue
                    logger.info(f"move to {(dest / path.name).absolute()}")
            if post_processor.get('run'):
                try:
                    process_output = subprocess.check_output(
                        post_processor['run'], shell=True,
                        input=reduce(lambda x, y: x + str(Path(y).absolute()) + '\n', file_list, ''),
                        stderr=subprocess.STDOUT, text=True)
                    logger.info(process_output.rstrip())
                except subprocess.CalledProcessError as e:
                    logger.exception(e.output)
                    continue



# 部分弹幕功能代码来自项目：https://github.com/IsoaSFlus/danmaku，感谢大佬
# 快手弹幕代码来源及思路：https://github.com/py-wuhao/ks_barrage，感谢大佬
# 部分斗鱼录播修复代码与思路来源于：https://github.com/SmallPeaches/DanmakuRender，感谢大佬
# 仅抓取用户弹幕，不包括入场提醒、礼物赠送等。

import asyncio
import os
import re
import ssl
import time
import logging
from typing import Optional

import lxml.etree as etree
import aiohttp

from biliup.plugins.Danmaku.douyu import Douyu
from biliup.plugins.Danmaku.huya import Huya
from biliup.plugins.Danmaku.bilibili import Bilibili
from biliup.plugins.Danmaku.twitch import Twitch
from biliup.plugins.Danmaku.douyin import Douyin

logger = logging.getLogger('biliup')


class DanmakuClient:
    class WebsocketErrorException(Exception):
        pass

    def __init__(self, url, filename):
        self.__starttime = time.time()
        self.__filename = os.path.splitext(filename)[0] + '.xml'
        self.__filename_video_suffix = filename
        self.__url = ''
        self.__site = None
        self.__hs = None
        self.__ws = None
        self.__dm_queue = None
        # 正在运行中的任务
        self.__danmaku_tasks: Optional[asyncio.Task] = None
        # 弹幕启动时的事件循环 为了保证被正确的关闭
        self.__loop = None

        if 'http://' == url[:7] or 'https://' == url[:8]:
            self.__url = url
        else:
            self.__url = 'http://' + url
        for u, s in {'douyu.com': Douyu,
                     'huya.com': Huya,
                     'live.bilibili.com': Bilibili,
                     'twitch.tv': Twitch,
                     'live.douyin.com': Douyin
                     }.items():
            if re.match(r'^(?:http[s]?://)?.*?%s/(.+?)$' % u, url):
                self.__site = s
                self.__u = u
                break

        if self.__site is None:
            logger.warning(f"{DanmakuClient.__name__}:{self.__url}:不支持录制弹幕")
            # 抛出异常由外部处理 exit()会导致进程退出
            raise Exception(f"{DanmakuClient.__name__}:{self.__url}:不支持录制弹幕")

    async def init_ws(self):
        try:
            ws_url, reg_datas = await self.__site.get_ws_info(self.__url)
            ctx = ssl.create_default_context()
            ctx.set_ciphers('DEFAULT')
            self.__ws = await self.__hs.ws_connect(ws_url, ssl_context=ctx, headers=getattr(self.__site, 'headers',{}))
            for reg_data in reg_datas:
                if type(reg_data) == str:
                    await self.__ws.send_str(reg_data)
                else:
                    await self.__ws.send_bytes(reg_data)
        except:
            raise self.WebsocketErrorException()

    async def heartbeats(self):
        while self.__site.heartbeat:
            # 每隔这么长时间发送一次心跳包
            await asyncio.sleep(self.__site.heartbeatInterval)
            # 发送心跳包
            if type(self.__site.heartbeat) == str:
                await self.__ws.send_str(self.__site.heartbeat)
            else:
                await self.__ws.send_bytes(self.__site.heartbeat)

    async def fetch_danmaku(self):
        while True:
            # 使用 async for msg in self.__ws
            # 会导致在连接断开时 需要很长时间15min或者更多才能检测到
            msg = await self.__ws.receive()

            if msg.type in [aiohttp.WSMsgType.CLOSED, aiohttp.WSMsgType.ERROR]:
                # 连接关闭的异常 会到外层统一处理
                raise self.WebsocketErrorException()

            try:
                result = self.__site.decode_msg(msg.data)

                if isinstance(result, tuple):
                    ms, ack = result
                    if ack is not None:
                        # 发送ack包
                        if type(ack) == str:
                            await self.__ws.send_str(ack)
                        else:
                            await self.__ws.send_bytes(ack)
                else:
                    ms = result

                for m in ms:
                    await self.__dm_queue.put(m)
            except Exception as Error:
                logger.warning(f"{DanmakuClient.__name__}:{self.__url}:弹幕接收异常：{Error}")
                continue
                # await asyncio.sleep(10) 无需等待 直接获取下一条websocket消息
                # 这里出现异常只会是 decode_msg 的问题

    async def print_danmaku(self):
        parser = etree.XMLParser(recover=True)
        root = etree.Element("root")
        tree = etree.ElementTree(root, parser=parser)
        msg_i = 0
        msg_col = {'0': '16777215', '1': '16717077', '2': '2000880', '3': '8046667', '4': '16744192',
                   '5': '10172916',
                   '6': '16738740'}

        def write_file(filename):
            try:
                with open(filename, "wb") as f:
                    etree.indent(root, "\t")
                    tree.write(f, encoding="UTF-8", xml_declaration=True, pretty_print=True)
            except Exception as Error:
                logger.warning(f"{DanmakuClient.__name__}:{self.__url}:弹幕写入异常：{Error}")

        try:
            while True:
                m = await self.__dm_queue.get()
                if m.get('msg_type') == 'danmaku':
                    try:
                        d = etree.SubElement(root, 'd')
                        if 'col' in m:
                            color = msg_col[m["col"]]
                        elif 'color' in m:
                            color = m["color"]
                        else:
                            color = '16777215'
                        msg_time = format(time.time() - self.__starttime, '.3f')
                        d.set('p', f"{msg_time},1,25,{color},0,0,0,0")
                        d.text = m["content"]
                    except Exception as Error:
                        logger.warning(f"{DanmakuClient.__name__}:{self.__url}:弹幕处理异常：{Error}")
                        # 异常后略过本次弹幕
                        continue

                    if msg_i >= 5:
                        # 收到五条弹幕后写入 减少io
                        # 可能会写入失败 会在下次五条或者任务被取消时重新尝试写入
                        write_file(self.__filename)
                        msg_i = 0
                    else:
                        msg_i = msg_i + 1
        except asyncio.CancelledError:
            # 在这段代码中只有执行m = await self.__dm_queue.get()的时候可能会被取消 所以无需担心其他问题
            # 任务被取消之前需要写入一下 不然会丢失没写入的
            write_file(self.__filename)
            # 如果不存在对应的视频则删除弹幕
            if not (os.path.exists(f"{self.__filename_video_suffix}.part") or
                    os.path.exists(f"{self.__filename_video_suffix}")):
                os.remove(self.__filename)

    async def start(self):
        self.__loop = asyncio.get_event_loop()
        self.__dm_queue = asyncio.Queue()
        is_first_error = True
        while True:
            try:
                self.__hs = aiohttp.ClientSession()
                await self.init_ws()
                # 保存以便取消
                self.__danmaku_tasks = asyncio.gather(self.heartbeats(), self.fetch_danmaku(), self.print_danmaku())
                await self.__danmaku_tasks
            except self.WebsocketErrorException:
                self.stop()
                if is_first_error:
                    # 减少日志输出量
                    logger.warning(f"{DanmakuClient.__name__}:{self.__url}:弹幕连接异常，将在 30 秒后重试")
                    is_first_error = False

                # 连接断开等30秒重连
                # 在关闭之前一直重试
                await asyncio.sleep(30)
                continue
            except asyncio.CancelledError:
                pass
            break

    def stop(self):
        # 关闭任务后loop就会被自动关闭 这里让他先保留一下 等待其他close完成后手动关闭
        sleep_task = asyncio.run_coroutine_threadsafe(asyncio.sleep(9999), self.__loop)
        # 关闭任务
        if self.__danmaku_tasks is not None:
            self.__danmaku_tasks.cancel()
        if self.__ws is not None:
            asyncio.run_coroutine_threadsafe(self.__ws.close(), self.__loop)
        if self.__hs is not None:
            asyncio.run_coroutine_threadsafe(self.__hs.close(), self.__loop)
        sleep_task.cancel()

# 虎牙直播：https://www.huya.com/lpl
# 斗鱼直播：https://www.douyu.com/9999



import requests

from biliup.config import config
from ..engine.decorators import Plugin
from ..plugins import match1, logger
from ..engine.download import DownloadBase

from urllib.parse import urlparse, parse_qs

@Plugin.download(regexp=r'(?:https?://)?(?:(?:live|www|v)\.)?(kuaishou)\.com')
@Plugin.download(regexp=r'(?:https?://)?(?:(?:(?:livev)\.(?:m))\.)?chenzhongtech\.com')
class Kuaishou(DownloadBase):
    def __init__(self, fname, url, suffix='flv'):
        super().__init__(fname, url, suffix)

    def check_stream(self, is_check=False):
        murl = ""
        nurl = f"https://live.kuaishou.com/u/{get_kwaiId(self.url)}"

        '''
        Api
        '''
        # split_args = ["/profile/", "/fw/live/"]
        # for split_key in split_args:
        #     if split_key in self.url:
        #         self.url = f"https://live.kuaishou.com/u/{self.url.split(split_key)[1]}"
        # from urllib.parse import urlparse
        # r = requests.get(self.url, headers=self.fake_headers)
        # parsed_url = urlparse(r.url)
        # kwaiId = parsed_url.path.split('/')[-1]
        # kwai_data = requests.get(f'https://live.kuaishou.com/live_api/profile/public?principalId={kwaiId}',
        #              headers=self.fake_headers).json()
        # live_data = kwai_data['data']['live']
        # # 被风控时会返回未开播
        # if not live_data['living']:
        #     logger.debug(kwaiId + '未开播')
        #     return False
        # self.room_title = live_data['caption']
        # self.raw_stream_url = live_data['playUrls'][0]['adaptationSet']['representation'][-1]['url']


        '''
        Phone_WEB
        '''
        # 可以输入 快手号(kwaiId), 用户唯一辨识ID(userEid), v.kuaishou.com 短链
        # livev.m.chenzhongtech.com 移动端链接也需要刷新验证参数
        headers = self.fake_headers.copy()
        headers['User-Agent'] = 'Mozilla/5.0 (iPhone; CPU iPhone OS 13_2_3 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.0.3 Mobile/15E148 Safari/604.1'
        with requests.Session() as s:
            try:
                s.headers.update(headers)
                r = s.get(nurl, allow_redirects=False)
                murl = r.headers['Location']
            except Exception:
                formatted_log("error", MURL=murl, NURL=nurl, OURL=self.url)
                raise
            parsed_url = urlparse(murl)
            eid = parsed_url.path.split('/')[-1]
            kpn = parse_qs(parsed_url.query).get('kpn')[0]
            s.headers.update({
                'Origin': 'https://livev.m.chenzhongtech.com',
                'Referer': murl,
                'Content-Type': 'application/json',
            })
            data = {
                'source': 6,
                'eid': eid, # 写作 eid 实际是 kwaiId
                'shareMethod': 'card',
                'clientType': 'WEB_OUTSIDE_SHARE_H5'
            }
            # 我不好说，但距离上面获取移动端页面的请求必须间隔 1s 以上
            import time
            time.sleep(1)
            live_data = s.post(f"https://livev.m.chenzhongtech.com/rest/k/live/byUser?kpn={kpn}",
                                json=data).json()
        if not live_data['result'] == 1 :
            formatted_log("error", BANNED=live_data['error_msg'])
            return False
        liveStream = live_data.get('liveStream')
        if liveStream is None:
            formatted_log("error", BANNED=live_data['error_msg'])
            return False
        # liveStream['type'] != 2 or liveStream['streamType'] != 1 可能是视频轮播，未发现相关主播
        if not liveStream['living']:
            logger.debug(liveStream['user']['user_name'] + "未开播")
            return False
        self.room_title = liveStream['caption']
        if config.get('kwai_protocol', "FLV").lower() == 'hls':
            PlayUrlInfo = liveStream['multiResolutionHlsPlayUrls'][-1]
            PlayUrls = PlayUrlInfo['urls'][0]
            self.raw_stream_url = PlayUrls['url']
        elif config.get('kwai_cdn') is not None:
            PlayUrlInfo = liveStream['multiResolutionPlayUrls'][-1]
            PlayUrls = PlayUrlInfo['urls']
            for playUrlInfo in PlayUrls:
                if config.get('kwai_cdn').lower() in playUrlInfo['cdn'].lower():
                    playUrl = playUrlInfo
                    self.raw_stream_url = playUrl['url']

        if self.raw_stream_url is None:
            import random
            liveAdaptiveManifest = random.choice(liveStream['liveAdaptiveManifest'])
            PlayUrlInfo = liveAdaptiveManifest['adaptationSet']['representation'][-1]
            self.raw_stream_url = PlayUrlInfo['url']
        else:
            streamName = self.raw_stream_url.split('/gifshow/')[1].split('.')[0]
            formatted_log('', QUALITY=PlayUrlInfo['type'], QUALITY_NAME=PlayUrlInfo['name'], LEVEL=PlayUrlInfo['level'],
                        URLTYPE=PlayUrls['urlType'], STREAMNAME=streamName)

        '''
        PC_WEB
        '''
        # with requests.Session() as s:
        #     if "/profile/" in self.url:
        #         self.url = f"https://live.kuaishou.com/u/{self.url.split('/profile/')[1]}"
        #     res = s.get(self.url, timeout=5, headers=self.fake_headers)
        # initial_state = res.text.split('window.__INITIAL_STATE__=')[1].split(';(')[0]
        # liveroom = json.loads(initial_state)['liveroom']
        # if liveroom['errorType']['type'] != 1:
        #     logger.debug(liveroom['errorType']['title'])
        #     return False
        # liveStream = liveroom['liveStream']
        # if not liveroom['isLiving'] or liveStream['type'] not in 'live':
        #     logger.debug("直播间未开播或播放的不是直播")
        #     return False
        # self.raw_stream_url = liveStream['playUrls'][0]['adaptationSet']['representation'][-1]['url']
        # self.room_title = liveStream['caption']
        # author = liveroom['author']
        # if self.use_live_cover is True:
        #     try:
        #         self.live_cover_path = \
        #         super().get_live_cover(author['name'], \
        #                                author['id'], \
        #                                self.room_title, \
        #                                author['timestamp'], \
        #                                liveStream['coverUrl'])
        #     except:
        #         logger.error(f"获取直播封面失败")
        return True

def formatted_log(*args, **kwargs):
    level = args[0] if args else ''
    for key, value in kwargs.items():
        if level == "error":
            logger.error(f"Kuaishou {key}: {value}")
        elif level == "warning":
            logger.warning(f"Kuaishou {key}: {value}")
        elif level == "info":
            logger.info(f"Kuaishou {key}: {value}")
        else:
            logger.debug(f"Kuaishou {key}: {value}")

def get_kwaiId(url):
    split_args = ["/profile/", "/fw/live/", "/u/"]
    for key in split_args:
        if key in url:
            kwaiId = url.split(key)[1]
            return kwaiId


import json
import re
from struct import pack

import aiohttp


class Douyu:
    wss_url = 'wss://danmuproxy.douyu.com:8506/'
    heartbeat = b"\x14\x00\x00\x00\x14\x00\x00\x00\xb1\x02\x00\x00\x74\x79\x70\x65\x40\x3d\x6d\x72\x6b\x6c\x2f\x00"
    heartbeatInterval = 30

    @staticmethod
    async def get_ws_info(url):
        room_id = url.split('/')[-1]
        async with aiohttp.ClientSession() as session:
            async with session.get('https://m.douyu.com/' + str(room_id)) as resp:
                room_page = await resp.text()
                room_id = re.findall(r'"rid":(\d{1,8})', room_page)[0]
        reg_datas = []
        data = f'type@=loginreq/roomid@={room_id}/'
        s = pack('i', 9 + len(data)) * 2
        s += b'\xb1\x02\x00\x00'  # 689
        s += data.encode('ascii') + b'\x00'
        reg_datas.append(s)
        data = f'type@=joingroup/rid@={room_id}/gid@=-9999/'
        s = pack('i', 9 + len(data)) * 2
        s += b'\xb1\x02\x00\x00'  # 689
        s += data.encode('ascii') + b'\x00'
        reg_datas.append(s)
        return Douyu.wss_url, reg_datas

    @staticmethod
    def decode_msg(data):
        msgs = []
        for msg in re.findall(b'(type@=.*?)\x00', data):
            msga = {}
            try:
                msg = msg.replace(b'@=', b'":"').replace(b'/', b'","')
                msg = msg.replace(b'@A', b'@').replace(b'@S', b'/')
                msg = json.loads((b'{"' + msg[:-2] + b'}').decode('utf8', 'ignore'))
                msga['name'] = msg.get('nn', '')
                msga['content'] = msg.get('txt', '')
                msga['msg_type'] = {'dgb': 'gift', 'chatmsg': 'danmaku',
                                   'uenter': 'enter'}.get(msg['type'], 'other')
                msga['col'] = msg.get('col', '0')
                msgs.append(msga)
            except Exception as Error:
                # print(f"decode_msg：捕获到异常：{Error}")
                pass
        return msgs
